/*
 *    Linux/68k Frame Buffer Device Configuration
 *
 *     Copyright 1995 by Geert Uytterhoeven
 *                       (Geert.Uytterhoeven@cs.kuleuven.ac.be)
 *
 *    --------------------------------------------------------------------------
 *
 *    This file is subject to the terms and conditions of the GNU General Public
 *    License. See the file COPYING in the main directory of the Linux/68k
 *    distribution for more details.
 */


#define VERSION   "Linux/68k Frame Buffer Device Configuration " \
                  "Version 0.2 (5/12/95)\n"  \
                  "(C) Copyright 1995 by Geert Uytterhoeven\n"


#include <stdarg.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <sys/ioctl.h>

struct file;
struct inode;

#include <linux/fb.h>

#include "fbset.h"


   /*
    *    Default Frame Buffer Special Device Node
    */

#define FB_DEVICE          "/dev/fb0current"


   /*
    *    Default Video Mode Database File
    */

#define FB_MODEDB          "/etc/fb.modes"


   /*
    *    Broadcast (PAL/NTSC) Visible Display Lines
    */

#define PAL_WINDOW_V       (288)
#define NTSC_WINDOW_V      (242)


   /*
    *    Command Line Options
    */

char *Opt_fb = FB_DEVICE;
char *Opt_db = FB_MODEDB;
char *Opt_xres = NULL;
char *Opt_yres = NULL;
char *Opt_vxres = NULL;
char *Opt_vyres = NULL;
char *Opt_depth = NULL;
char *Opt_pixclock = NULL;
char *Opt_left = NULL;
char *Opt_right = NULL;
char *Opt_upper = NULL;
char *Opt_lower = NULL;
char *Opt_hslen = NULL;
char *Opt_vslen = NULL;
char *Opt_hsync = NULL;
char *Opt_vsync = NULL;
char *Opt_csync = NULL;
char *Opt_extsync = NULL;
char *Opt_bcast = NULL;
char *Opt_laced = NULL;
char *Opt_double = NULL;
char *Opt_move = NULL;
char *Opt_step = NULL;

char *Opt_mode = NULL;

struct {
   char *name;
   char **value;
} Options[] = {
   { "fb", &Opt_fb },
   { "db", &Opt_db },
   { "xres", &Opt_xres },
   { "yres", &Opt_yres },
   { "vxres", &Opt_vxres },
   { "vyres", &Opt_vyres },
   { "depth", &Opt_depth },
   { "pixclock", &Opt_pixclock },
   { "left", &Opt_left },
   { "right", &Opt_right },
   { "upper", &Opt_upper },
   { "lower", &Opt_lower },
   { "hslen", &Opt_hslen },
   { "vslen", &Opt_vslen },
   { "hsync", &Opt_hsync },
   { "vsync", &Opt_vsync },
   { "csync", &Opt_csync },
   { "extsync", &Opt_extsync },
   { "bcast", &Opt_bcast },
   { "laced", &Opt_laced },
   { "double", &Opt_double },
   { "move", &Opt_move },
   { "step", &Opt_step },
   { NULL, NULL }
};

char *ProgramName;

u_short Verbose = 0;
u_short Force = 0;
u_short Info = 0;
u_short Compare = 0;
u_short Now = 0;


   /*
    *    Monitor Specifications
    */

struct MonSpecs MonSpecs;


   /*
    *    Video Mode Database
    */

struct VideoMode *VideoModes = NULL;


   /*
    *    Current Video Mode
    */

struct VideoMode Current;
struct VideoMode Changed;


   /*
    *    Function Prototypes
    */

static void Usage(void);
static void ConvertFromVideoMode(const struct VideoMode *vmode,
                                 struct fb_var_screeninfo *var);
static void ConvertToVideoMode(const struct fb_var_screeninfo *var,
                               struct VideoMode *vmode);
static void CompareVideoModes(struct VideoMode *a, struct VideoMode *b);
static u_long atopolarity(const char *var, const char *name);
static u_long atoboolean(const char *var, const char *name);
#if 0
static void ReadVideoModes(void);
#endif
static struct VideoMode *FindVideoMode(const char *name);
static void ModifyVideoMode(struct VideoMode *vmode);
static void DisplayVModeInfo(struct VideoMode *vmode);
static void DisplayFBInfo(int fh);
static int SafeVideoMode(const struct VideoMode *vmode);
int main(int argc, char *argv[]);


   /*
    *    Print an Error Message and Exit
    */

void Die(const char *fmt, ...)
{
   va_list ap;

   fflush(stdout);
   va_start(ap, fmt);
   vfprintf(stderr, fmt, ap);
   va_end(ap);
   exit(1);
}


   /*
    *    Print the Usage Template and Exit
    */

static void Usage(void)
{
   Die("Usage: %s [options] [mode]\n\n"
       "Valid options:\n"
       "   General options:\n"
       "      --help            : display this usage information\n"
       "      --verbose         : verbose mode\n"
       "      --force           : force unsafe mode change\n"
       "      --info            : only display info, do not change video mode\n"
       "      --compare         : display changes made by frame buffer device\n"
       "      --now             : switch to new mode immediately\n"
       "                          (default for " FB_DEVICE ")\n"
       "   Frame buffer device:\n"
       "      -fb <device>      : frame buffer special device node\n"
       "                          (default is " FB_DEVICE ")\n"
       "   Video mode database:\n"
       "      -db <file>        : video mode database file\n"
       "                          (default is " FB_MODEDB ")\n"
       "   Display geometry:\n"
       "      -xres <value>     : horizontal resolution (in pixels)\n"
       "      -yres <value>     : vertical resolution (in pixels)\n"
       "      -vxres <value>    : virtual horizontal resolution (in pixels)\n"
       "      -vyres <value>    : virtual vertical resolution (in pixels)\n"
       "      -depth <value>    : display depth (in bits per pixel)\n"
       "   Display timings:\n"
       "      -pixclock <value> : pixel clock (in picoseconds)\n"
       "      -left <value>     : left margin (in pixels)\n"
       "      -right <value>    : right margin (in pixels)\n"
       "      -upper <value>    : upper margin (in pixel lines)\n"
       "      -lower <value>    : lower margin (in pixel lines)\n"
       "      -hslen <value>    : horizontal sync length (in pixels)\n"
       "      -vslen <value>    : vertical sync length (in pixel lines)\n"
       "   Display flags:\n"
       "      -hsync <value>    : horizontal sync polarity (low or high)\n"
       "      -vsync <value>    : vertical sync polarity (low or high)\n"
       "      -csync <value>    : composite sync polarity (low or high)\n"
       "      -extsync <value>  : external sync enable (false or true)\n"
       "      -bcast <value>    : broadcast enable (false or true)\n"
       "      -laced <value>    : interlace enable (false or true)\n"
       "      -double <value>   : doublescan enable (false or true)\n"
       "   Display positioning:\n"
       "      -move <direction> : move the visible part (left, right, up or down)\n"
       "      -step <value>     : step increment (in pixels or pixel lines)\n"
       "                          (default is 8 horizontal, 2 vertical)\n",
       ProgramName);
}


   /*
    *    Conversion Routines
    */

static void ConvertFromVideoMode(const struct VideoMode *vmode,
                                 struct fb_var_screeninfo *var)
{
   memset(var, 0, sizeof(struct fb_var_screeninfo));
   var->xres = vmode->geometry.xres;
   var->yres = vmode->geometry.yres;
   var->xres_virtual = vmode->geometry.vxres;
   var->yres_virtual = vmode->geometry.vyres;
   var->bits_per_pixel = vmode->geometry.depth;
   if (Now || !strcmp(Opt_fb, FB_DEVICE))
      var->activate = FB_ACTIVATE_NOW;
   else
      var->activate = FB_ACTIVATE_TEST;
   var->pixclock = vmode->timings.pixclock;
   var->left_margin = vmode->timings.left;
   var->right_margin = vmode->timings.right;
   var->upper_margin = vmode->timings.upper;
   var->lower_margin = vmode->timings.lower;
   var->hsync_len = vmode->timings.hslen;
   var->vsync_len = vmode->timings.vslen;
   if (vmode->options.hsync == HIGH)
      var->sync |= FB_SYNC_HOR_HIGH_ACT;
   if (vmode->options.vsync == HIGH)
      var->sync |= FB_SYNC_VERT_HIGH_ACT;
   if (vmode->options.csync == HIGH)
      var->sync |= FB_SYNC_COMP_HIGH_ACT;
   if (vmode->options.extsync == TRUE)
      var->sync |= FB_SYNC_EXT;
   if (vmode->options.bcast == TRUE)
      var->sync |= FB_SYNC_BROADCAST;
   if (vmode->options.laced == TRUE)
      var->vmode = FB_VMODE_INTERLACED;
   else if (vmode->options.dblscan == TRUE)
      var->vmode = FB_VMODE_DOUBLE;
   else
      var->vmode = FB_VMODE_NONINTERLACED;
}


static void ConvertToVideoMode(const struct fb_var_screeninfo *var,
                               struct VideoMode *vmode)
{
   vmode->name = NULL;
   vmode->geometry.xres = var->xres;
   vmode->geometry.yres = var->yres;
   vmode->geometry.vxres = var->xres_virtual;
   vmode->geometry.vyres = var->yres_virtual;
   vmode->geometry.depth = var->bits_per_pixel;
   vmode->timings.pixclock = var->pixclock;
   vmode->timings.left = var->left_margin;
   vmode->timings.right = var->right_margin;
   vmode->timings.upper = var->upper_margin;
   vmode->timings.lower = var->lower_margin;
   vmode->timings.hslen = var->hsync_len;
   vmode->timings.vslen = var->vsync_len;
   vmode->options.hsync = var->sync & FB_SYNC_HOR_HIGH_ACT ? HIGH : LOW;
   vmode->options.vsync = var->sync & FB_SYNC_VERT_HIGH_ACT ? HIGH : LOW;
   vmode->options.csync = var->sync & FB_SYNC_COMP_HIGH_ACT ? HIGH : LOW;
   vmode->options.extsync = var->sync & FB_SYNC_EXT ? TRUE : FALSE;
   vmode->options.bcast = var->sync & FB_SYNC_BROADCAST ? TRUE : FALSE;
   vmode->options.laced = FALSE;
   vmode->options.dblscan = FALSE;
   switch (var->vmode & FB_VMODE_MASK) {
      case FB_VMODE_INTERLACED:
         vmode->options.laced = TRUE;
         break;
      case FB_VMODE_DOUBLE:
         vmode->options.dblscan = TRUE;
         break;
   }
}


static void CompareVideoModes(struct VideoMode *a, struct VideoMode *b)
{
   u_short changed = 0;

   if (a->geometry.xres != b->geometry.xres) {
      printf("   xres changed from %d to %d\n", a->geometry.xres,
             b->geometry.xres);
      changed = 1;
   }
   if (a->geometry.yres != b->geometry.yres) {
      printf("   yres changed from %d to %d\n", a->geometry.yres,
             b->geometry.yres);
      changed = 1;
   }
   if (a->geometry.vxres != b->geometry.vxres) {
      printf("   vxres changed from %d to %d\n", a->geometry.vxres,
             b->geometry.vxres);
      changed = 1;
   }
   if (a->geometry.vyres != b->geometry.vyres) {
      printf("   vyres changed from %d to %d\n", a->geometry.vyres,
             b->geometry.vyres);
      changed = 1;
   }
   if (a->geometry.depth != b->geometry.depth) {
      printf("   depth changed from %d to %d\n", a->geometry.depth,
             b->geometry.depth);
      changed = 1;
   }
   if (a->timings.pixclock != b->timings.pixclock) {
      printf("   pixclock changed from %d to %d\n", a->timings.pixclock,
             b->timings.pixclock);
      changed = 1;
   }
   if (a->timings.left != b->timings.left) {
      printf("   left changed from %d to %d\n", a->timings.left,
             b->timings.left);
      changed = 1;
   }
   if (a->timings.right != b->timings.right) {
      printf("   right changed from %d to %d\n", a->timings.right,
             b->timings.right);
      changed = 1;
   }
   if (a->timings.upper != b->timings.upper) {
      printf("   upper changed from %d to %d\n", a->timings.upper,
             b->timings.upper);
      changed = 1;
   }
   if (a->timings.lower != b->timings.lower) {
      printf("   lower changed from %d to %d\n", a->timings.lower,
             b->timings.lower);
      changed = 1;
   }
   if (a->timings.hslen != b->timings.hslen) {
      printf("   hslen changed from %d to %d\n", a->timings.hslen,
             b->timings.hslen);
      changed = 1;
   }
   if (a->timings.vslen != b->timings.vslen) {
      printf("   vslen changed from %d to %d\n", a->timings.vslen,
             b->timings.vslen);
      changed = 1;
   }
   if (a->options.hsync != b->options.hsync) {
      printf("   hsync changed from %s to %s\n",
             a->options.hsync ? "high" : "low",
             b->options.hsync ? "high" : "low");
      changed = 1;
   }
   if (a->options.vsync != b->options.vsync) {
      printf("   vsync changed from %s to %s\n",
             a->options.vsync ? "high" : "low",
             b->options.vsync ? "high" : "low");
      changed = 1;
   }
   if (a->options.csync != b->options.csync) {
      printf("   csync changed from %s to %s\n",
             a->options.csync ? "high" : "low",
             b->options.csync ? "high" : "low");
      changed = 1;
   }
   if (a->options.extsync != b->options.extsync) {
      printf("   extsync changed from %s to %s\n",
             a->options.extsync ? "true" : "false",
             b->options.extsync ? "true" : "false");
      changed = 1;
   }
   if (a->options.bcast != b->options.bcast) {
      printf("   bcast changed from %s to %s\n",
             a->options.bcast ? "true" : "false",
             b->options.bcast ? "true" : "false");
      changed = 1;
   }
   if (a->options.laced != b->options.laced) {
      printf("   laced changed from %s to %s\n",
             a->options.laced ? "true" : "false",
             b->options.laced ? "true" : "false");
      changed = 1;
   }
   if (a->options.dblscan != b->options.dblscan) {
      printf("   dblscan changed from %s to %s\n",
             a->options.dblscan ? "true" : "false",
             b->options.dblscan ? "true" : "false");
      changed = 1;
   }
   if (a->options.hsync != b->options.hsync) {
      printf("   hsync changed from %s to %s\n",
             a->options.hsync ? "true" : "false",
             b->options.hsync ? "true" : "false");
      changed = 1;
   }
   if (!changed)
      puts("   nothing changed by frame buffer device");
}


static u_long atopolarity(const char *var, const char *name)
{
   u_long value = 0;

   if (!strcasecmp(var, "low"))
      value = 0;
   else if (!strcasecmp(var, "high"))
      value = 1;
   else
      Die("The %s polarity must be `low' or `high'\n", name);

   return(value);
}


static u_long atoboolean(const char *var, const char *name)
{
   u_long value = 0;

   if (!strcasecmp(var, "false"))
      value = 0;
   else if (!strcasecmp(var, "true"))
      value = 1;
   else
      Die("The %s flag must be `false' or `true'\n", name);

   return(value);
}


void AddVideoMode(struct VideoMode *vmode)
{
   struct VideoMode *vm2;

   if (FindVideoMode(vmode->name))
      Die("%s:%d: Duplicate mode name `%s'\n", Opt_db, line, vmode->name);
   vm2 = malloc(sizeof(struct VideoMode));
   *vm2 = *vmode;
   vm2->next = VideoModes;
   VideoModes = vm2;
}


   /*
    *    Read the Video Mode Database
    */

static void ReadVideoModes(void)
{
   if (!(yyin = fopen(Opt_db, "r")))
      Die("fopen %s: %s\n", Opt_db, strerror(errno));
   yyparse();
   fclose(yyin);
}


   /*
    *    Find a Video Mode
    */

static struct VideoMode *FindVideoMode(const char *name)
{
   struct VideoMode *vmode;
   int i;

   for (vmode = VideoModes; vmode; vmode = vmode->next)
      if (!strcasecmp(name, vmode->name))
         break;

   return(vmode);
}


   /*
    *    Modify a Video Mode
    */

static void ModifyVideoMode(struct VideoMode *vmode)
{
   u_long step;

   if (Opt_xres)
      vmode->geometry.xres = atol(Opt_xres);
   if (Opt_yres)
      vmode->geometry.yres = atol(Opt_yres);
   if (Opt_vxres)
      vmode->geometry.vxres = atol(Opt_vxres);
   if (Opt_vyres)
      vmode->geometry.vyres = atol(Opt_vyres);
   if (Opt_depth)
      vmode->geometry.depth = atol(Opt_depth);
   if (Opt_pixclock)
      vmode->timings.pixclock = atol(Opt_pixclock);
   if (Opt_left)
      vmode->timings.left = atol(Opt_left);
   if (Opt_right)
      vmode->timings.right = atol(Opt_right);
   if (Opt_upper)
      vmode->timings.upper = atol(Opt_upper);
   if (Opt_lower)
      vmode->timings.lower = atol(Opt_lower);
   if (Opt_hslen)
      vmode->timings.hslen = atol(Opt_hslen);
   if (Opt_vslen)
      vmode->timings.vslen = atol(Opt_vslen);
   if (Opt_hsync)
      vmode->options.hsync = atopolarity(Opt_hsync, "horizontal sync polarity");
   if (Opt_vsync)
      vmode->options.vsync = atopolarity(Opt_vsync, "vertical sync polarity");
   if (Opt_csync)
      vmode->options.csync = atopolarity(Opt_csync, "composite sync polarity");
   if (Opt_extsync)
      vmode->options.extsync = atoboolean(Opt_extsync, "external sync enable");
   if (Opt_bcast)
      vmode->options.bcast = atoboolean(Opt_bcast, "broadcast enable");
   if (Opt_laced)
      vmode->options.laced = atoboolean(Opt_laced, "interlace enable");
   if (Opt_double)
      vmode->options.dblscan = atoboolean(Opt_double, "doublescan enable");
   if (Opt_step)
      step = atol(Opt_step);
   if (Opt_move)
      if (!strcasecmp(Opt_move, "left")) {
         if (!Opt_step)
            step = 8;
         if (step > vmode->timings.left)
            Die("The left margin cannot be negative\n");
         vmode->timings.left -= step;
         vmode->timings.right += step;
      } else if (!strcasecmp(Opt_move, "right")) {
         if (!Opt_step)
            step = 8;
         if (step > vmode->timings.right)
            Die("The right margin cannot be negative\n");
         vmode->timings.left += step;
         vmode->timings.right -= step;
      } else if (!strcasecmp(Opt_move, "up")) {
         if (!Opt_step)
            step = 2;
         if (step > vmode->timings.upper)
            Die("The upper margin cannot be negative\n");
         vmode->timings.upper -= step;
         vmode->timings.lower += step;
      } else if (!strcasecmp(Opt_move, "down")) {
         if (!Opt_step)
            step = 2;
         if (step > vmode->timings.lower)
            Die("The lower margin cannot be negative\n");
         vmode->timings.upper += step;
         vmode->timings.lower -= step;
      } else
         Die("The move direction must be `left', `right', `up' or `down'\n");
}


   /*
    *    Display some Video Mode Information
    */

static void DisplayVModeInfo(struct VideoMode *vmode)
{
   if (Verbose) {
      puts("\nmode \"<name>\"");
      puts("   geometry <xres> <yres> <vxres> <vyres> <depth>");
      puts("   timings <pixclock> <left> <right> <upper> <lower> <hslen> <vslen>");
      puts("   [hsync {low|high}]");
      puts("   [vsync {low|high}]");
      puts("   [csync {low|high}]");
      puts("   [extsync {false|true}]");
      puts("   [bcast {false|true}]");
      puts("   [laced {false|true}]");
      puts("   [double {false|true}]");
      puts("endmode\n");
   }
   printf("mode \"change_my_name\"\n");
   printf("   geometry %d %d %d %d %d\n", vmode->geometry.xres,
          vmode->geometry.yres, vmode->geometry.vxres, vmode->geometry.vyres,
          vmode->geometry.depth);
   printf("   timings %d %d %d %d %d %d %d\n", vmode->timings.pixclock,
          vmode->timings.left, vmode->timings.right, vmode->timings.upper,
          vmode->timings.lower, vmode->timings.hslen, vmode->timings.vslen);
   if (vmode->options.hsync)
      puts("   hsync high");
   if (vmode->options.vsync)
      puts("   vsync high");
   if (vmode->options.csync)
      puts("   csync high");
   if (vmode->options.extsync)
      puts("   extsync true");
   if (vmode->options.bcast)
      puts("   bcast true");
   if (vmode->options.laced)
      puts("   laced true");
   if (vmode->options.dblscan)
      puts("   double true");
   puts("endmode\n");
}


   /*
    *    Display some Frame Buffer Device Information
    */

static void DisplayFBInfo(int fh)
{
   struct fb_fix_screeninfo fix;

   if (ioctl(fh, FBIOGET_FSCREENINFO, &fix))
      Die("ioctl FBIOGET_FSCREENINFO: %s\n", strerror(errno));
   puts("Frame buffer device information:");
	printf("   Name     : %s\n", fix.id);
	printf("   Address  : 0x%08lx\n", fix.smem_start);
	printf("   Size     : %ld\n", fix.smem_len);
   printf("   Type     : ");
   switch (fix.type) {
      case FB_TYPE_PACKED_PIXELS:
         printf("PACKED PIXELS\n");
         break;
      case FB_TYPE_PLANES:
         printf("PLANES\n");
         break;
      case FB_TYPE_INTERLEAVED_PLANES:
         printf("INTERLEAVED PLANES (%d bytes interleave)\n", fix.type_aux);
         break;
      default:
         printf("%d (UNKNOWN) Type_aux %d\n", fix.type, fix.type_aux);
   }
   printf("   Visual   : ");
   switch (fix.visual) {
      case FB_VISUAL_MONO01:
         printf("MONO01\n");
         break;
      case FB_VISUAL_MONO10:
         printf("MONO10\n");
         break;
      case FB_VISUAL_TRUECOLOR:
         printf("TRUECOLOR\n");
         break;
      case FB_VISUAL_PSEUDOCOLOR:
         printf("PSEUDOCOLOR\n");
         break;
      case FB_VISUAL_DIRECTCOLOR:
         printf("DIRECTCOLOR\n");
         break;
      case FB_VISUAL_STATIC_PSEUDOCOLOR:
         printf("STATIC PSEUDOCOLOR\n");
         break;
      case FB_VISUAL_STATIC_DIRECTCOLOR:
         printf("STATIC DIRECTCOLOR\n");
         break;
      default:
         printf("%d (UNKNOWN)\n", fix.visual);
         break;
   }
   printf("   XPanStep : %d\n", fix.xpanstep);
   printf("   YPanStep : %d\n", fix.xpanstep);
   printf("   YWrapStep: %d\n", fix.ywrapstep);
	puts("");
}


   /*
    *    This routine should check whether a Video Mode is Safe for your Monitor
    *
    *    Of course I do _not_ guarantee it will protect your equipment!!
    */

static int SafeVideoMode(const struct VideoMode *vmode)
{
   int safe = 0;
   double hrate, vrate;
   u_short line_shift_incd;
   u_long htotal, vtotal;

   htotal = vmode->timings.left+vmode->geometry.xres+vmode->timings.right+
            vmode->timings.hslen;
   vtotal = vmode->timings.upper+vmode->geometry.yres+vmode->timings.lower+
            vmode->timings.vslen;

   if (vmode->options.laced)
      vtotal >>= 1;
   else if (vmode->options.dblscan)
      vtotal <<= 1;

   if (!htotal || !vtotal)
      Die("Empty video mode\n");

   if (vmode->options.bcast) {
      if (vmode->timings.hslen)
         Die("Hslen must be zero for a broadcast mode\n");
      if (vmode->timings.vslen)
         Die("Vslen must be zero for a broadcast mode\n");
      if (vtotal == PAL_WINDOW_V) {
         hrate = 15625;
         vrate = 50;
      } else if (vtotal == NTSC_WINDOW_V) {
         hrate = 15750;
         vrate = 60;
      } else
         Die("Bad values for a broadcast mode\n");
   } else {
      hrate = 1E12/(htotal*vmode->timings.pixclock);
      vrate = hrate/vtotal;
   }

   if ((hrate >= MonSpecs.MinHrate) && (hrate <= MonSpecs.MaxHrate) &&
       (vrate >= MonSpecs.MinVrate) && (vrate <= MonSpecs.MaxVrate))
      safe = 1;

   if (Verbose || Info)
      printf("Refresh: Horizontal = %5.2f kHz   Vertical = %5.2f Hz\n",
             hrate/1000.0, vrate);

   return(safe);
}


   /*
    *    Main Routine
    */

int main(int argc, char *argv[])
{
   struct VideoMode *vmode;
   struct fb_var_screeninfo var;
   int fh, flags, i;

   /*
    *    Parse the Options
    */

   ProgramName = argv[0];
   puts(VERSION);
   argc--;
   argv++;
   if (!argc)
      Usage();
   while (argc) {
      if (!strcmp(argv[0], "--help"))
         Usage();
      else if (!strcmp(argv[0], "--verbose"))
         Verbose = 1;
      else if (!strcmp(argv[0], "--force"))
         Force = 1;
      else if (!strcmp(argv[0], "--info"))
         Info = 1;
      else if (!strcmp(argv[0], "--compare"))
         Compare = 1;
      else if (!strcmp(argv[0], "--now"))
         Now = 1;
      else {
         for (i = 0; Options[i].name; i++)
            if (!strcmp(&argv[0][1], Options[i].name))
               break;
         if (Options[i].name)
            if (--argc) {
               *Options[i].value = argv[1];
               argv++;
            } else
               Usage();
         else if (argc == 1)
            Opt_mode = argv[0];
         else
            Usage();
      }
      argc--;
      argv++;
   }

   /*
    *    Open the Frame Buffer Device
    */

#if 0
   if (Now || !strcmp(Opt_fb, FB_DEVICE))
      flags = O_RDWR;
   else
#else
      flags = O_RDONLY;
#endif

   if ((fh = open(Opt_fb, flags)) == -1)
      Die("open %s: %s\n", Opt_fb, strerror(errno));

   /*
    *    Read the Video Mode Database
    */

   ReadVideoModes();

   /*
    *    Get the Video Mode
    */

   if (Opt_mode)
      if (!(vmode = FindVideoMode(Opt_mode)))
         Die("Unknown video mode `%s'\n", Opt_mode);     
      else {
         Current = *vmode;
         if (Verbose)
            printf("Using video mode `%s'\n", Opt_mode);
      }
   else {
      if (ioctl(fh, FBIOGET_VSCREENINFO, &var))
         Die("ioctl FBIOGET_VSCREENINFO: %s\n", strerror(errno));
      ConvertToVideoMode(&var, &Current);
      if (Verbose)
         puts("Using current video mode");
   }

   /*
    *    Optionally Modify the Video Mode
    */

   ModifyVideoMode(&Current);

   /*
    *    Display some Video Mode Information
    */

   if (Info || Verbose) {
      DisplayVModeInfo(&Current);
      if (Verbose)
         DisplayFBInfo(fh);
   }

   /*
    *    Check whether the Video Mode is Safe
    */

   if (!SafeVideoMode(&Current)) {
      puts("\n*** WARNING: This video mode may damage your monitor!! ***\n");
      if (!Force && !Info)
         Die("             Use option -f to force continue\n");
   }

   /*
    *    Set the Video Mode
    */

   if (!Info) {
      ConvertFromVideoMode(&Current, &var);
      if (ioctl(fh, FBIOPUT_VSCREENINFO, &var))
         Die("ioctl FBIOPUT_VSCREENINFO: %s\n", strerror(errno));
      if (Compare) {
         if (ioctl(fh, FBIOGET_VSCREENINFO, &var))
            Die("ioctl FBIOGET_VSCREENINFO: %s\n", strerror(errno));
         ConvertToVideoMode(&var, &Changed);
         CompareVideoModes(&Current, &Changed);
      }
   }

   /*
    *    Close the Frame Buffer Device
    */

   close(fh);

   exit(0);
}
